@@ -2,6 +2,7 @@
  * ntfsclone - Part of the Linux-NTFS project.
  *
  * Copyright (c) 2003-2004 Szabolcs Szakacsits
+ * Copyright (c) 2004 Anton Altaparmakov
  *
  * Clone NTFS data and/or metadata to a sparse file, device or stdout.
  *
@@ -77,13 +78,12 @@ struct progress_bar {
 	float unit;
 };
 
-struct __ntfs_walk_clusters_ctx {
+typedef struct {
 	ntfs_inode *ni;			/* inode being processed */
 	ntfs_attr_search_ctx *ctx;	/* inode attribute being processed */
 	s64 inuse;			/* number of clusters in use */
-};
+} ntfs_walk_clusters_ctx;
 
-typedef struct __ntfs_walk_clusters_ctx ntfs_walk_clusters_ctx;
 typedef int (ntfs_walk_op)(ntfs_inode *ni, void *data);
 
 struct ntfs_walk_cluster {
@@ -120,17 +120,10 @@ int wiped_timestamp_data = 0;
 #define read_all(f, p, n)  io_all((f), (p), (n), 0)
 #define write_all(f, p, n) io_all((f), (p), (n), 1)
 
-/* FIXME: They should be #included but Eprintf conflicts with mkntfs's Eprintf */
-extern int Eprintf (const char *format, ...) __attribute__ ((format (printf, 1, 2)));
-extern int Vprintf (const char *format, ...) __attribute__ ((format (printf, 1, 2)));
-extern int Qprintf (const char *format, ...) __attribute__ ((format (printf, 1, 2)));
-/* FIXME: should be 'static' */
-extern int Printf (const char *format, ...) __attribute__ ((format (printf, 1, 2)));
-
 GEN_PRINTF(Eprintf, stderr,  NULL,         FALSE)
 GEN_PRINTF(Vprintf, msg_out, &opt.verbose, TRUE)
 GEN_PRINTF(Qprintf, msg_out, &opt.quiet,   FALSE)
-GEN_PRINTF(Printf,  msg_out, NULL,         FALSE)
+static GEN_PRINTF(Printf,  msg_out, NULL,         FALSE)
 
 
 static void perr_printf(const char *fmt, ...)
@@ -250,7 +243,7 @@ static void parse_options(int argc, char **argv)
 		case 'o':
 			if (opt.output)
 				usage();
-			opt.output = argv[optind-1];
+			opt.output = optarg;
 			break;
 		default:
 			err_printf("Unknown option '%s'.\n", argv[optind-1]);
@@ -349,18 +342,29 @@ static s64 nr_clusters_to_bitmap_byte_size(s64 nr_clusters)
 	return bm_bsize;
 }
 
-static int is_critical_metadata(ntfs_walk_clusters_ctx *image)
+static s64 is_critical_metadata(ntfs_walk_clusters_ctx *image, runlist *rl)
 {
-	s64 inode;
-	
-	inode = image->ni->mft_no;
+	s64 inode = image->ni->mft_no;
 
-	if (inode <= LAST_METADATA_INODE)
+	if (inode <= LAST_METADATA_INODE) {
 		if (inode != FILE_LogFile)
-			return 1;
+			return rl->length;
+
+		if (image->ctx->attr->type == AT_DATA) {
+			/* Save at least the first 8 KiB of FILE_LogFile */
+			s64 s = (s64)8192 - rl->vcn * vol->cluster_size;
+			if (s > 0) {
+				s = rounded_up_division(s, vol->cluster_size);
+				if (rl->length < s)
+					s = rl->length;
+				return s;
+			}
+			return 0;
+		}
+	}
 	
 	if (image->ctx->attr->type != AT_DATA)
-		return 1;
+		return rl->length;
 
 	return 0;
 }
@@ -425,18 +429,18 @@ static void lseek_to_cluster(s64 lcn)
 
 static void dump_clusters(ntfs_walk_clusters_ctx *image, runlist *rl)
 {
-	int i;
+	s64 i, len; /* number of clusters to copy */
 	
-	if (opt.std_out)
+	if (opt.std_out || !opt.metadata_only)
 		return;
 	
-	if (!opt.metadata_only || !is_critical_metadata(image))
+	if (!(len = is_critical_metadata(image, rl)))
 		return;
-
+	
 	lseek_to_cluster(rl->lcn);
 	
 	/* FIXME: this could give pretty suboptimal performance */
-	for (i = 0; i < rl->length; i++)
+	for (i = 0; i < len; i++)
 		copy_cluster();
 }
 
@@ -523,10 +527,10 @@ static void wipe_resident_data(ntfs_walk_clusters_ctx *image)
 	ATTR_RECORD *a;
 	u32 i;
 	int n = 0;
-	char *p;
+	u8 *p;
 
 	a = image->ctx->attr;
-	p = (char *)a + a->value_offset;
+	p = (u8*)a + le16_to_cpu(a->value_offset);
 	
 	if (image->ni->mft_no <= LAST_METADATA_INODE)
 		return;
@@ -534,7 +538,7 @@ static void wipe_resident_data(ntfs_walk_clusters_ctx *image)
 	if (a->type != AT_DATA)
 		return;
 	
-	for (i = 0; i < a->value_length; i++) {
+	for (i = 0; i < le32_to_cpu(a->value_length); i++) {
 		if (p[i]) {
 			p[i] = 0;
 			n++;
@@ -697,8 +701,9 @@ static void wipe_unused_mft_data(ntfs_inode *ni)
 	if (ni->mft_no <= LAST_METADATA_INODE)
 		return;
 	
-	unused = m->bytes_allocated - m->bytes_in_use;
-	wiped_unused_mft_data += wipe_data((char *)m, m->bytes_in_use, unused);
+	unused = le32_to_cpu(m->bytes_allocated) - le32_to_cpu(m->bytes_in_use);
+	wiped_unused_mft_data += wipe_data((char *)m,
+			le32_to_cpu(m->bytes_in_use), unused);
 }
 
 static void wipe_unused_mft(ntfs_inode *ni)
@@ -710,7 +715,7 @@ static void wipe_unused_mft(ntfs_inode *ni)
 	if (ni->mft_no <= LAST_METADATA_INODE)
 		return;
 	
-	unused = m->bytes_in_use - sizeof(MFT_RECORD);
+	unused = le32_to_cpu(m->bytes_in_use) - sizeof(MFT_RECORD);
 	wiped_unused_mft += wipe_data((char *)m, sizeof(MFT_RECORD), unused);
 }
 
@@ -779,7 +784,7 @@ static int walk_clusters(ntfs_volume *volume, struct ntfs_walk_cluster *walk)
 		if (wipe)
 			nr_used_mft_records++;
 
-		if ((ni->mrec->base_mft_record) != 0)
+		if (ni->mrec->base_mft_record)
 			goto out;
 
 		walk->image->ni = ni;
@@ -906,7 +911,8 @@ static void mount_volume(unsigned long new_mntflag)
 	if (ntfs_version_is_supported(vol))
 		perr_exit("Unknown NTFS version");
 
-	Printf("Cluster size       : %u bytes\n", vol->cluster_size);
+	Printf("Cluster size       : %u bytes\n",
+			(unsigned int)vol->cluster_size);
 	print_volume_size("Current volume size",
 			  volume_size(vol, vol->nr_clusters));
 }
@@ -1046,7 +1052,8 @@ int main(int argc, char **argv)
 	       if ((fd_out = fileno(stdout)) == -1) 
 		       perr_exit("fileno for stdout failed");
 	} else {
-		int flags = O_WRONLY;
+	        /* device_size_get() might need to read() */
+		int flags = O_RDWR;
 		
 		if (!opt.blkdev_out) {
 			flags |= O_CREAT | O_TRUNC;
@@ -1121,4 +1128,3 @@ int main(int argc, char **argv)
 	fsync_clone(fd_out);
 	exit(0);
 }
-
