@@ -2,7 +2,7 @@
  * ntfsresize - Part of the Linux-NTFS project.
  *
  * Copyright (c) 2002-2004 Szabolcs Szakacsits
- * Copyright (c) 2002-2003 Anton Altaparmakov
+ * Copyright (c) 2002-2004 Anton Altaparmakov
  * Copyright (c) 2002-2003 Richard Russon
  *
  * This utility will resize an NTFS volume without data loss.
@@ -107,20 +107,29 @@ struct llcn_t {
 	s64 inode;	/* inode using it */
 };	
 	
-struct __ntfs_resize_t {
-	s64 new_volume_size;		/* in clusters; 0 = --info w/o --size */
-	int shrink;			/* shrink = 1, enlarge = 0 */
-	ntfs_inode *ni;			/* inode being processed */
-	MFT_REF mref;                   /* mft reference */
-	MFT_RECORD *mrec;               /* mft record */
-	ntfs_attr_search_ctx *ctx;	/* inode attribute being processed */
-	u64 relocations;		/* num of clusters to relocate */
-	s64 inuse;			/* num of clusters in use */
-	int multi_ref;			/* num of clusters ref'd many times */
-	runlist mftmir_rl;		/* $MFTMirr AT_DATA's new position */
-	s64 mftmir_old;			/* $MFTMirr AT_DATA's old LCN */
-	int dirty_inode;		/* some inode data got relocated */
+typedef struct {
+	ntfs_inode *ni;		     /* inode being processed */
+	ntfs_attr_search_ctx *ctx;   /* inode attribute being processed */
+	s64 inuse;		     /* num of clusters in use */
+	int multi_ref;		     /* num of clusters referenced many times */
+	struct bitmap lcn_bitmap;
+} ntfs_fsck_t;
+
+typedef struct {
+	ntfs_volume *vol;
+	s64 new_volume_size;	     /* in clusters; 0 = --info w/o --size */
+	int shrink;		     /* shrink = 1, enlarge = 0 */
+	ntfs_inode *ni;		     /* inode being processed */
+	MFT_REF mref;                /* mft reference */
+	MFT_RECORD *mrec;            /* mft record */
+	ntfs_attr_search_ctx *ctx;   /* inode attribute being processed */
+	u64 relocations;	     /* num of clusters to relocate */
+	s64 inuse;		     /* num of clusters in use */
+	runlist mftmir_rl;	     /* $MFTMirr AT_DATA's new position */
+	s64 mftmir_old;		     /* $MFTMirr AT_DATA's old LCN */
+	int dirty_inode;	     /* some inode data got relocated */
 	struct progress_bar progress;
+	struct bitmap lcn_bitmap;
 	/* Temporary statistics until all case is supported */
 	struct llcn_t last_mft;
 	struct llcn_t last_mftmir;
@@ -128,18 +137,11 @@ struct __ntfs_resize_t {
 	struct llcn_t last_sparse;
 	struct llcn_t last_compressed;
 	struct llcn_t last_lcn;
-	s64 last_unsafe;		/* last "unsafe-to-move" cluster */
-	s64 last_unsupp;		/* last unsupported cluster */
-};
+	s64 last_unsupp;	     /* last unsupported cluster */
+} ntfs_resize_t;
 
-typedef struct __ntfs_resize_t ntfs_resize_t;
-
-/* FIXME: This will go to struct __ntfs_resize_t */
-ntfs_volume *vol = NULL;
-
-/* FIXME: These and 'pos' from find_free_cluster() will make a cluster
-   allocation related structure, attached to struct __ntfs_resize_t */
-struct bitmap lcn_bitmap;
+/* FIXME: This, lcn_bitmap and pos from find_free_cluster() will make a cluster
+   allocation related structure, attached to ntfs_resize_t */
 s64 max_free_cluster_range = 0;
 
 #define NTFS_MBYTE (1000 * 1000)
@@ -157,14 +159,9 @@ s64 max_free_cluster_range = 0;
 
 #define rounded_up_division(a, b) (((a) + (b - 1)) / (b))
 
-/* FIXME: They should be included but Eprintf conflicts with mkntfs's Eprintf */
-extern int Eprintf(const char *format, ...) __attribute__ ((format (printf, 1, 2)));
-extern int Vprintf(const char *format, ...) __attribute__ ((format (printf, 1, 2)));
-extern int Qprintf(const char *format, ...) __attribute__ ((format (printf, 1, 2)));
-
 GEN_PRINTF(Eprintf, stderr, NULL,         FALSE)
 GEN_PRINTF(Vprintf, stdout, &opt.verbose, TRUE)
-GEN_PRINTF(Qprintf, stdout, NULL,   FALSE)
+GEN_PRINTF(Qprintf, stdout, NULL,         FALSE)
 
 /**
  * perr_printf
@@ -303,7 +300,7 @@ static void version (void)
 {
 	printf ("\nResize an NTFS Volume, without data loss.\n\n");
 	printf ("Copyright (c) 2002-2004  Szabolcs Szakacsits\n");
-	printf ("Copyright (c) 2002-2003  Anton Altaparmakov\n");
+	printf ("Copyright (c) 2002-2004  Anton Altaparmakov\n");
 	printf ("Copyright (c) 2002-2003  Richard Russon\n");
 	printf ("\n%s\n%s%s", ntfs_gpl, ntfs_bugs, ntfs_home);
 }
@@ -478,7 +475,7 @@ static int parse_options(int argc, char **argv)
 	return (!err && !help && !ver);
 }
 
-static void print_advise(s64 supp_lcn)
+static void print_advise(ntfs_volume *vol, s64 supp_lcn)
 {
 	s64 old_b, new_b, freed_b, old_mb, new_mb, freed_mb;
 	
@@ -558,10 +555,7 @@ static s64 nr_clusters_to_bitmap_byte_size(s64 nr_clusters)
 	s64 bm_bsize;
 
 	bm_bsize = rounded_up_division(nr_clusters, 8);
-
 	bm_bsize = (bm_bsize + 7) & ~7;
-	Dprintf("Bitmap byte size  : %lld (%lld clusters)\n",
-	       bm_bsize, rounded_up_division(bm_bsize, vol->cluster_size));
 
 	return bm_bsize;
 }
@@ -592,7 +586,7 @@ static int has_bad_sectors(ntfs_resize_t *resize)
 		return -1;
 
 	if (ustr && ntfs_names_are_equal(ustr, len,
-			(uchar_t*)((char*)a + le16_to_cpu(a->name_offset)),
+			(uchar_t*)((u8*)a + le16_to_cpu(a->name_offset)),
 			a->name_length, 0, NULL, 0)) 
 		ret = 1;
 									
@@ -602,7 +596,7 @@ static int has_bad_sectors(ntfs_resize_t *resize)
 	return ret;	
 }
 
-static void collect_shrink_constraints(ntfs_resize_t *resize, runlist *rl)
+static void collect_resize_constraints(ntfs_resize_t *resize, runlist *rl)
 {
 	s64 inode, last_lcn;
 	ATTR_FLAGS flags;
@@ -660,9 +654,6 @@ static void collect_shrink_constraints(ntfs_resize_t *resize, runlist *rl)
 		llcn->inode = inode;
 	}	
 
-	if (resize->last_unsafe < last_lcn)
-		resize->last_unsafe = last_lcn;
-
 	if (supported)
 		return;
 
@@ -671,7 +662,7 @@ static void collect_shrink_constraints(ntfs_resize_t *resize, runlist *rl)
 }
 
 
-static void collect_shrink_info(ntfs_resize_t *resize, runlist *rl)
+static void collect_relocation_info(ntfs_resize_t *resize, runlist *rl)
 {
 	s64 lcn, lcn_length, start, len, inode;
 	s64 new_vol_size;	/* (last LCN on the volume) + 1 */
@@ -697,7 +688,7 @@ static void collect_shrink_info(ntfs_resize_t *resize, runlist *rl)
 			err_printf("$MFT%s can't be split up yet. Please try "
 				   "a different size.\n", inode ? "Mirr" : "");
 			last_lcn = lcn + lcn_length - 1;
-			print_advise(last_lcn);
+			print_advise(resize->vol, last_lcn);
 			exit(1);
 		}
 	}
@@ -709,8 +700,8 @@ static void collect_shrink_info(ntfs_resize_t *resize, runlist *rl)
 	
 	printf("Relocation needed for inode %8lld attr 0x%x LCN 0x%08llx "
 			"length %6lld\n", (long long)inode,
-			resize->ctx->attr->type, (unsigned long long)start,
-			(long long)len);
+			(unsigned int)le32_to_cpu(resize->ctx->attr->type),
+			(unsigned long long)start, (long long)len);
 }
 
 /**
@@ -722,15 +713,16 @@ static void collect_shrink_info(ntfs_resize_t *resize, runlist *rl)
  *
  * This serves as a rudimentary "chkdsk" operation.
  */
-static void build_lcn_usage_bitmap(ntfs_resize_t *resize)
+static void build_lcn_usage_bitmap(ntfs_volume *vol, ntfs_fsck_t *fsck)
 {
 	s64 inode;
 	ATTR_RECORD *a;
 	runlist *rl;
 	int i, j;
+	struct bitmap *lcn_bitmap = &fsck->lcn_bitmap;
 
-	a = resize->ctx->attr;
-	inode = resize->ni->mft_no;
+	a = fsck->ctx->attr;
+	inode = fsck->ni->mft_no;
 
 	if (!a->non_resident)
 		return;
@@ -756,13 +748,13 @@ static void build_lcn_usage_bitmap(ntfs_resize_t *resize)
 		if (lcn < 0 || lcn_length <= 0)
 			err_exit("Corrupt runlist in inode %lld attr %x LCN "
 				 "%llx length %llx\n", inode,
-				 le32_to_cpu (a->type), lcn, lcn_length);
+				 le32_to_cpu(a->type), lcn, lcn_length);
 
 		for (j = 0; j < lcn_length; j++) {
 			u64 k = (u64)lcn + j;
-			if (ntfs_bit_get_and_set(lcn_bitmap.bm, k, 1)) {
+			if (ntfs_bit_get_and_set(lcn_bitmap->bm, k, 1)) {
 				
-				if (++resize->multi_ref > 10)
+				if (++fsck->multi_ref > 10)
 					continue;
 
 				printf("Cluster %llu (0x%llx) referenced "
@@ -771,13 +763,7 @@ static void build_lcn_usage_bitmap(ntfs_resize_t *resize)
 						(unsigned long long)k);
 			}
 		}
-
-		resize->inuse += lcn_length;
-		
-		collect_shrink_constraints(resize, rl + i);
-		
-		if (resize->shrink)
-			collect_shrink_info(resize, rl + i);
+		fsck->inuse += lcn_length;
 	}
 	free(rl);
 }
@@ -788,21 +774,18 @@ static void build_lcn_usage_bitmap(ntfs_resize_t *resize)
  * For a given MFT Record, iterate through all its attributes.  Any non-resident
  * data runs will be marked in lcn_bitmap.
  */
-static void walk_attributes(ntfs_resize_t *resize)
+static void walk_attributes(ntfs_volume *vol, ntfs_fsck_t *fsck)
 {
-	ntfs_attr_search_ctx *ctx;
-
-	if (!(ctx = ntfs_attr_get_search_ctx(resize->ni, NULL)))
+	if (!(fsck->ctx = ntfs_attr_get_search_ctx(fsck->ni, NULL)))
 		perr_exit("ntfs_get_attr_search_ctx");
 
-	while (!ntfs_attrs_walk(ctx)) {
-		if (ctx->attr->type == AT_END)
+	while (!ntfs_attrs_walk(fsck->ctx)) {
+		if (fsck->ctx->attr->type == AT_END)
 			break;
-		resize->ctx = ctx;
-		build_lcn_usage_bitmap(resize);
+		build_lcn_usage_bitmap(vol, fsck);
 	}
 
-	ntfs_attr_put_search_ctx(ctx);
+	ntfs_attr_put_search_ctx(fsck->ctx);
 }
 
 /**
@@ -811,7 +794,7 @@ static void walk_attributes(ntfs_resize_t *resize)
  * Compare two bitmaps.  In this case, $Bitmap as read from the disk and
  * lcn_bitmap which we built from the MFT Records.
  */
-static void compare_bitmaps(struct bitmap *a)
+static void compare_bitmaps(ntfs_volume *vol, struct bitmap *a)
 {
 	s64 i, pos, count;
 	int mismatch = 0;
@@ -918,20 +901,19 @@ static void progress_update(struct progress_bar *p, u64 current)
  * Read each record in the MFT, skipping the unused ones, and build up a bitmap
  * from all the non-resident attributes.
  */
-static void walk_inodes(ntfs_resize_t *resize)
+static void build_allocation_bitmap(ntfs_volume *vol, ntfs_fsck_t *fsck)
 {
 	s64 inode = 0;
-	s64 last_mft_rec;
 	ntfs_inode *ni;
+	struct progress_bar progress;
 
 	/* WARNING: don't modify the text, external tools grep for it */
 	printf("Checking filesystem consistency ...\n");
 
-	last_mft_rec = vol->nr_mft_records - 1;
-	progress_init(&resize->progress, inode, last_mft_rec, 100);
+	progress_init(&progress, inode, vol->nr_mft_records - 1, 100);
 
-	for (; inode <= last_mft_rec; inode++) {
-		progress_update(&resize->progress, inode);
+	for (; inode < vol->nr_mft_records; inode++) {
+		progress_update(&progress, inode);
 
 		if ((ni = ntfs_inode_open(vol, (MFT_REF)inode)) == NULL) {
 			/* FIXME: continue only if it make sense, e.g.
@@ -941,11 +923,76 @@ static void walk_inodes(ntfs_resize_t *resize)
 			perr_exit("Reading inode %lld failed", inode);
 		}
 
-		if ((ni->mrec->base_mft_record) != 0)
+		if (ni->mrec->base_mft_record)
+			goto close_inode;
+
+		fsck->ni = ni;
+		walk_attributes(vol, fsck);
+close_inode:
+		if (ntfs_inode_close(ni))
+			perr_exit("ntfs_inode_close for inode %lld", inode);
+	}
+}
+
+static void build_resize_constrains(ntfs_resize_t *resize)
+{
+	s64 i;
+	runlist *rl;
+
+	if (!resize->ctx->attr->non_resident)
+		return;
+
+	if (!(rl = ntfs_mapping_pairs_decompress(resize->vol, 
+						 resize->ctx->attr, NULL)))
+		perr_exit("ntfs_decompress_mapping_pairs");
+
+	for (i = 0; rl[i].length; i++) {
+		/* CHECKME: LCN_RL_NOT_MAPPED check isn't needed */
+		if (rl[i].lcn == LCN_HOLE || rl[i].lcn == LCN_RL_NOT_MAPPED)
+			continue;
+		
+		collect_resize_constraints(resize, rl + i);
+		if (resize->shrink)
+			collect_relocation_info(resize, rl + i);
+	}
+	free(rl);
+}
+
+static void resize_constrains_by_attributes(ntfs_resize_t *resize)
+{
+	if (!(resize->ctx = ntfs_attr_get_search_ctx(resize->ni, NULL)))
+		perr_exit("ntfs_get_attr_search_ctx");
+
+	while (!ntfs_attrs_walk(resize->ctx)) {
+		if (resize->ctx->attr->type == AT_END)
+			break;
+		build_resize_constrains(resize);
+	}
+
+	ntfs_attr_put_search_ctx(resize->ctx);
+}
+
+static void set_resize_constrains(ntfs_resize_t *resize)
+{
+	s64 inode;
+	ntfs_inode *ni;
+
+	printf("Collecting shrinkage constrains ...\n");
+
+	for (inode = 0; inode < resize->vol->nr_mft_records; inode++) {
+
+		ni = ntfs_inode_open(resize->vol, (MFT_REF)inode);
+		if (ni == NULL) {
+			if (errno == EIO || errno == ENOENT)
+				continue;
+			perr_exit("Reading inode %lld failed", inode);
+		}
+
+		if (ni->mrec->base_mft_record)
 			goto close_inode;
 
 		resize->ni = ni;
-		walk_attributes(resize);
+		resize_constrains_by_attributes(resize);
 close_inode:
 		if (ntfs_inode_close(ni))
 			perr_exit("ntfs_inode_close for inode %lld", inode);
@@ -981,7 +1028,9 @@ static void rl_fixup(runlist **rl)
 	}
 }
 
-static void replace_attribute_runlist(ntfs_attr_search_ctx *ctx, runlist *rl)
+static void replace_attribute_runlist(ntfs_volume *vol, 
+				      ntfs_attr_search_ctx *ctx, 
+				      runlist *rl)
 {
 	int mp_size, l;
 	void *mp;
@@ -1014,28 +1063,33 @@ static void replace_attribute_runlist(ntfs_attr_search_ctx *ctx, runlist *rl)
 		
 		Vprintf("Old mp size      : %d\n", l);
 		Vprintf("New mp size      : %d\n", mp_size);
-		Vprintf("Bytes in use     : %d\n", ctx->mrec->bytes_in_use);
+		Vprintf("Bytes in use     : %u\n", (unsigned int)
+				le32_to_cpu(ctx->mrec->bytes_in_use));
 	
 		next_attr = (char *)a + le16_to_cpu(a->length);
 		l = mp_size - l;
 		
-		Vprintf("Bytes in use new : %d\n", ctx->mrec->bytes_in_use + l);
-		Vprintf("Bytes allocated  : %d\n", ctx->mrec->bytes_allocated);
+		Vprintf("Bytes in use new : %u\n", l + (unsigned int)
+				le32_to_cpu(ctx->mrec->bytes_in_use));
+		Vprintf("Bytes allocated  : %u\n", (unsigned int)
+				le32_to_cpu(ctx->mrec->bytes_allocated));
 
-		remains_size = ctx->mrec->bytes_in_use;
+		remains_size = le32_to_cpu(ctx->mrec->bytes_in_use);
 		remains_size -= (next_attr - (char *)ctx->mrec);
 
 		Vprintf("increase         : %d\n", l);
 		Vprintf("shift            : %lld\n", (long long)remains_size);
 		
-		if (ctx->mrec->bytes_in_use + l > ctx->mrec->bytes_allocated)
+		if (le32_to_cpu(ctx->mrec->bytes_in_use) + l >
+				le32_to_cpu(ctx->mrec->bytes_allocated))
 			err_exit("Extended record needed (%d > %d), not yet "
 				 "supported!\nPlease try to free less space.\n",
-				 ctx->mrec->bytes_in_use + l,
-				 ctx->mrec->bytes_allocated);
+				 le32_to_cpu(ctx->mrec->bytes_in_use) + l,
+				 le32_to_cpu(ctx->mrec->bytes_allocated));
 
 		memmove(next_attr + l, next_attr, remains_size);
-		ctx->mrec->bytes_in_use += l;
+		ctx->mrec->bytes_in_use = cpu_to_le32(l +
+				le32_to_cpu(ctx->mrec->bytes_in_use));
 		a->length += l;
 	}
 
@@ -1045,27 +1099,27 @@ static void replace_attribute_runlist(ntfs_attr_search_ctx *ctx, runlist *rl)
 	if (ntfs_mapping_pairs_build(vol, mp, mp_size, rl))
 		perr_exit("ntfs_mapping_pairs_build");
 
-	memmove((char *)a + le16_to_cpu(a->mapping_pairs_offset), mp, mp_size);
+	memmove((u8*)a + le16_to_cpu(a->mapping_pairs_offset), mp, mp_size);
 
 	free(mp);
 }
 
-static void set_bitmap_range(s64 pos, s64 length, u8 bit)
+static void set_bitmap_range(struct bitmap *bm, s64 pos, s64 length, u8 bit)
 {
 	while (length--)
-		ntfs_bit_set(lcn_bitmap.bm, pos++, bit);
+		ntfs_bit_set(bm->bm, pos++, bit);
 }
 
-static void set_bitmap_clusters(runlist *rl, u8 bit)
+static void set_bitmap_clusters(struct bitmap *bm, runlist *rl, u8 bit)
 {
 	for (; rl->length; rl++)
-		set_bitmap_range(rl->lcn, rl->length, bit);
+		set_bitmap_range(bm, rl->lcn, rl->length, bit);
 }
 
-static void release_bitmap_clusters(runlist *rl)
+static void release_bitmap_clusters(struct bitmap *bm, runlist *rl)
 {
 	max_free_cluster_range = 0;
-	set_bitmap_clusters(rl, 0);
+	set_bitmap_clusters(bm, rl, 0);
 }
 
 static void set_max_free_zone(s64 length, s64 end, runlist_element *rle)
@@ -1076,7 +1130,10 @@ static void set_max_free_zone(s64 length, s64 end, runlist_element *rle)
 	}
 }
 
-static int find_free_cluster(runlist_element *rle, s64 nr_vol_clusters, int hint)
+static int find_free_cluster(struct bitmap *bm, 
+			     runlist_element *rle,
+			     s64 nr_vol_clusters, 
+			     int hint)
 {
 	/* FIXME: get rid of this 'static' variable */
 	static s64 pos = 0;
@@ -1093,7 +1150,7 @@ static int find_free_cluster(runlist_element *rle, s64 nr_vol_clusters, int hint
 	i = pos;
 
 	do {
-		if (!ntfs_bit_get(lcn_bitmap.bm, i)) {
+		if (!ntfs_bit_get(bm->bm, i)) {
 			if (++free_zone == items) {
 				set_max_free_zone(free_zone, i + 1, rle);
 				break;
@@ -1126,11 +1183,14 @@ static int find_free_cluster(runlist_element *rle, s64 nr_vol_clusters, int hint
 	if (pos == nr_vol_clusters)
 		pos = 0;
 
-	set_bitmap_range(rle->lcn, rle->length, 1);
+	set_bitmap_range(bm, rle->lcn, rle->length, 1);
 	return 0;
 }
 
-static runlist *alloc_cluster(s64 items, s64 nr_vol_clusters, int hint)
+static runlist *alloc_cluster(struct bitmap *bm, 
+			      s64 items, 
+			      s64 nr_vol_clusters, 
+			      int hint)
 {
 	runlist_element rle;
 	runlist *rl = NULL;
@@ -1147,7 +1207,7 @@ static runlist *alloc_cluster(s64 items, s64 nr_vol_clusters, int hint)
 		if (runs)
 			hint = 0;
 		rle.length = items;
-		if (find_free_cluster(&rle, nr_vol_clusters, hint) == -1)
+		if (find_free_cluster(bm, &rle, nr_vol_clusters, hint) == -1)
 			return NULL;
 		
 		rl_size = (runs + 2) * sizeof(runlist_element);
@@ -1219,18 +1279,18 @@ static int write_all(struct ntfs_device *dev, void *buf, int count)
  * Write an MFT Record back to the disk.  If the read-only command line option
  * was given, this function will do nothing.
  */
-static int write_mft_record(const MFT_REF mref, MFT_RECORD *mrec)
+static int write_mft_record(ntfs_volume *v, const MFT_REF mref, MFT_RECORD *buf)
 {
-	if (ntfs_mft_record_write(vol, mref, mrec))
+	if (ntfs_mft_record_write(v, mref, buf))
 		perr_exit("ntfs_mft_record_write");
 
-//	if (vol->dev->d_ops->sync(vol->dev) == -1)
+//	if (v->dev->d_ops->sync(v->dev) == -1)
 //		perr_exit("Failed to sync device");
 
 	return 0;
 }
 
-static void lseek_to_cluster(s64 lcn)
+static void lseek_to_cluster(ntfs_volume *vol, s64 lcn)
 {
 	off_t pos;
 	
@@ -1244,15 +1304,16 @@ static void copy_clusters(ntfs_resize_t *resize, s64 dest, s64 src, s64 len)
 {
 	s64 i;
 	char buff[NTFS_MAX_CLUSTER_SIZE]; /* overflow checked at mount time */
-	
+	ntfs_volume *vol = resize->vol;
+
 	for (i = 0; i < len; i++) {
 
-		lseek_to_cluster(src + i);
+		lseek_to_cluster(vol, src + i);
 	
 		if (read_all(vol->dev, buff, vol->cluster_size) == -1)
 			perr_exit("read_all");
 		
-		lseek_to_cluster(dest + i);
+		lseek_to_cluster(vol, dest + i);
 	
 		if (write_all(vol->dev, buff, vol->cluster_size) == -1)
 			perr_exit("write_all");
@@ -1361,15 +1422,16 @@ static void relocate_run(ntfs_resize_t *resize, runlist **rl, int run)
 	}
 
 	hint = (resize->mref == FILE_MFTMirr) ? 1 : 0;
-	if (!(relocate_rl = alloc_cluster(lcn_length, new_vol_size, hint)))
+	if (!(relocate_rl = alloc_cluster(&resize->lcn_bitmap, lcn_length, 
+					  new_vol_size, hint)))
 		perr_exit("Cluster allocation failed for %llu:%lld",
 			  resize->mref, lcn_length);
 	
 	/* FIXME: check $MFTMirr DATA isn't multi-run (or support it) */
 	Vprintf("Relocate inode %7llu:0x%x:%08lld:0x%08llx --> 0x%08llx\n",
 			(unsigned long long)resize->mref,
-			resize->ctx->attr->type, (long long)lcn_length,
-			(unsigned long long)lcn,
+			(unsigned int)le32_to_cpu(resize->ctx->attr->type),
+			(long long)lcn_length, (unsigned long long)lcn,
 			(unsigned long long)relocate_rl->lcn);
 
 	relocate_clusters(resize, relocate_rl, lcn);
@@ -1393,7 +1455,7 @@ static void relocate_attribute(ntfs_resize_t *resize)
 	if (!a->non_resident)
 		return;
 
-	if (!(rl = ntfs_mapping_pairs_decompress(vol, a, NULL)))
+	if (!(rl = ntfs_mapping_pairs_decompress(resize->vol, a, NULL)))
 		perr_exit("ntfs_decompress_mapping_pairs");
 
 	for (i = 0; rl[i].length; i++) {
@@ -1413,7 +1475,7 @@ static void relocate_attribute(ntfs_resize_t *resize)
 	}
 
 	if (resize->dirty_inode == DIRTY_ATTRIB) {
-		replace_attribute_runlist(resize->ctx, rl);
+		replace_attribute_runlist(resize->vol, resize->ctx, rl);
 		resize->dirty_inode = DIRTY_INODE;
 	}
 
@@ -1422,24 +1484,21 @@ static void relocate_attribute(ntfs_resize_t *resize)
 
 static void relocate_attributes(ntfs_resize_t *resize)
 {
-	ntfs_attr_search_ctx *ctx;
-
-	if (!(ctx = ntfs_attr_get_search_ctx(NULL, resize->mrec)))
+	if (!(resize->ctx = ntfs_attr_get_search_ctx(NULL, resize->mrec)))
 		perr_exit("ntfs_get_attr_search_ctx");
 
-	while (!ntfs_attrs_walk(ctx)) {
-		if (ctx->attr->type == AT_END)
+	while (!ntfs_attrs_walk(resize->ctx)) {
+		if (resize->ctx->attr->type == AT_END)
 			break;
-		resize->ctx = ctx;
 		relocate_attribute(resize);
 	}
 
-	ntfs_attr_put_search_ctx(ctx);
+	ntfs_attr_put_search_ctx(resize->ctx);
 }
 
 static void relocate_inode(ntfs_resize_t *resize, MFT_REF mref)
 {
-	if (ntfs_file_record_read(vol, mref, &resize->mrec, NULL)) {
+	if (ntfs_file_record_read(resize->vol, mref, &resize->mrec, NULL)) {
 		/* FIXME: continue only if it make sense, e.g.
 		   MFT record not in use based on $MFT bitmap */
 		if (errno == EIO || errno == ENOENT)
@@ -1458,7 +1517,7 @@ static void relocate_inode(ntfs_resize_t *resize, MFT_REF mref)
 	if (resize->dirty_inode == DIRTY_INODE) {
 //		if (vol->dev->d_ops->sync(vol->dev) == -1)
 //			perr_exit("Failed to sync device");
-		if (write_mft_record(mref, resize->mrec))
+		if (write_mft_record(resize->vol, mref, resize->mrec))
 			perr_exit("Couldn't update inode %llu", mref);
 	}
 }
@@ -1472,11 +1531,11 @@ static void relocate_inodes(ntfs_resize_t *resize)
 	progress_init(&resize->progress, 0, resize->relocations, 100);
 	resize->relocations = 0;
 	
-	resize->mrec = (MFT_RECORD *)malloc(vol->mft_record_size);
+	resize->mrec = (MFT_RECORD *)malloc(resize->vol->mft_record_size);
 	if (!resize->mrec)
 		perr_exit("malloc failed");
 
-   	for (mref = 1; mref < (MFT_REF)vol->nr_mft_records; mref++)
+   	for (mref = 1; mref < (MFT_REF)resize->vol->nr_mft_records; mref++)
 		relocate_inode(resize, mref);
 
 	relocate_inode(resize, 0);
@@ -1485,7 +1544,7 @@ static void relocate_inodes(ntfs_resize_t *resize)
 		free(resize->mrec);
 }
 
-static void print_hint(const char *s, struct llcn_t llcn)
+static void print_hint(ntfs_volume *vol, const char *s, struct llcn_t llcn)
 {
 	s64 runs_b, runs_mb;
 	
@@ -1507,19 +1566,21 @@ static void print_hint(const char *s, struct llcn_t llcn)
  */
 static void advise_on_resize(ntfs_resize_t *resize)
 {
+	ntfs_volume *vol = resize->vol;
+
 	printf("Estimating smallest shrunken size supported ...\n");
 
 	printf("File feature         Last used at      By inode\n");
-	print_hint("$MFT", resize->last_mft);
-	print_hint("Multi-Record", resize->last_multi_mft);
+	print_hint(vol, "$MFT", resize->last_mft);
+	print_hint(vol, "Multi-Record", resize->last_multi_mft);
 	if (opt.verbose) {
-		print_hint("$MFTMirr", resize->last_mftmir);
-		print_hint("Compressed", resize->last_compressed);
-		print_hint("Sparse", resize->last_sparse);
-		print_hint("Ordinary", resize->last_lcn);
+		print_hint(vol, "$MFTMirr", resize->last_mftmir);
+		print_hint(vol, "Compressed", resize->last_compressed);
+		print_hint(vol, "Sparse", resize->last_sparse);
+		print_hint(vol, "Ordinary", resize->last_lcn);
 	}
 
-	print_advise(resize->last_unsupp);
+	print_advise(vol, resize->last_unsupp);
 }
 
 /**
@@ -1542,10 +1603,12 @@ static void bitmap_file_data_fixup(s64 cluster, struct bitmap *bm)
  * FIXME: this function should go away and instead using a generalized
  * "truncate_bitmap_data_attr()"
  */
-static void truncate_badclust_bad_attr(ntfs_attr_search_ctx *ctx, s64 nr_clusters)
+static void truncate_badclust_bad_attr(ntfs_resize_t *resize)
 {
 	runlist *rl_bad;
-	ATTR_RECORD *a = ctx->attr;
+	ATTR_RECORD *a = resize->ctx->attr;
+	s64 nr_clusters = resize->new_volume_size;
+	ntfs_volume *vol = resize->vol;
 
 	if (!a->non_resident)
 		/* FIXME: handle resident attribute value */
@@ -1561,7 +1624,7 @@ static void truncate_badclust_bad_attr(ntfs_attr_search_ctx *ctx, s64 nr_cluster
 	rl_set(rl_bad, 0LL, (LCN)LCN_HOLE, nr_clusters);
 	rl_set(rl_bad + 1, nr_clusters, -1LL, 0LL);
 
-	replace_attribute_runlist(ctx, rl_bad);
+	replace_attribute_runlist(vol, resize->ctx, rl_bad);
 
 	free(rl_bad);
 }
@@ -1572,7 +1635,10 @@ static void truncate_badclust_bad_attr(ntfs_attr_search_ctx *ctx, s64 nr_cluster
  * Shrink the metadata file $Bitmap.  It must be large enough for one bit per
  * cluster of the shrunken volume.  Also it must be a of 8 bytes in size.
  */
-static void shrink_bitmap_data_attr(runlist **rlist, s64 nr_bm_clusters, s64 new_size)
+static void shrink_bitmap_data_attr(struct bitmap *bm, 
+				    runlist **rlist, 
+				    s64 nr_bm_clusters, 
+				    s64 new_size)
 {
 	runlist *rl = *rlist;
 	int i, j;
@@ -1593,7 +1659,7 @@ static void shrink_bitmap_data_attr(runlist **rlist, s64 nr_bm_clusters, s64 new
 
 			k = rl[i].lcn + j;
 			if (k < new_size) {
-				ntfs_bit_set(lcn_bitmap.bm, k, 0);
+				ntfs_bit_set(bm->bm, k, 0);
 				Dprintf("Unallocate cluster: "
 				       "%lld (%llx)\n", k, k);
 			}
@@ -1617,17 +1683,21 @@ static void shrink_bitmap_data_attr(runlist **rlist, s64 nr_bm_clusters, s64 new
  * Enlarge the metadata file $Bitmap.  It must be large enough for one bit per
  * cluster of the shrunken volume.  Also it must be a of 8 bytes in size.
  */
-static void enlarge_bitmap_data_attr(runlist **rl, s64 nr_bm_clusters, s64 new_size)
+static void enlarge_bitmap_data_attr(ntfs_volume *vol,
+				     struct bitmap *bm,
+				     runlist **rl, 
+				     s64 nr_bm_clusters, 
+				     s64 new_size)
 {
 	s64 i;
 
-	release_bitmap_clusters(*rl);
+	release_bitmap_clusters(bm, *rl);
 	free(*rl);
 
 	for (i = vol->nr_clusters; i < new_size; i++)
-		ntfs_bit_set(lcn_bitmap.bm, i, 0);
+		ntfs_bit_set(bm->bm, i, 0);
 
-	if (!(*rl = alloc_cluster(nr_bm_clusters, new_size, 0)))
+	if (!(*rl = alloc_cluster(bm, nr_bm_clusters, new_size, 0)))
 		perr_exit("Couldn't allocate $Bitmap clusters");
 }
 
@@ -1642,6 +1712,7 @@ static void truncate_bitmap_data_attr(ntfs_resize_t *resize)
 	s64 nr_bm_clusters;
 	s64 nr_clusters;
 	u8 *tmp;
+	ntfs_volume *vol = resize->vol;
 
 	a = resize->ctx->attr;
 	if (!a->non_resident)
@@ -1652,11 +1723,11 @@ static void truncate_bitmap_data_attr(ntfs_resize_t *resize)
 	bm_bsize = nr_clusters_to_bitmap_byte_size(nr_clusters);
 	nr_bm_clusters = rounded_up_division(bm_bsize, vol->cluster_size);
 
-	if (!(tmp = (u8 *)realloc(lcn_bitmap.bm, bm_bsize)))
+	if (!(tmp = (u8 *)realloc(resize->lcn_bitmap.bm, bm_bsize)))
 		perr_exit("realloc");
-	lcn_bitmap.bm = tmp;
-	lcn_bitmap.size = bm_bsize;
-	bitmap_file_data_fixup(nr_clusters, &lcn_bitmap);
+	resize->lcn_bitmap.bm = tmp;
+	resize->lcn_bitmap.size = bm_bsize;
+	bitmap_file_data_fixup(nr_clusters, &resize->lcn_bitmap);
 
 	if (!(rl = ntfs_mapping_pairs_decompress(vol, a, NULL)))
 		perr_exit("ntfs_mapping_pairs_decompress");
@@ -1664,23 +1735,25 @@ static void truncate_bitmap_data_attr(ntfs_resize_t *resize)
 	/* NOTE: shrink could use enlarge_bitmap_data_attr() also. Advantages:
 	   less code, better code coverage. "Drawback": could be relocated */
 	if (resize->shrink)
-		shrink_bitmap_data_attr(&rl, nr_bm_clusters, nr_clusters);
+		shrink_bitmap_data_attr(&resize->lcn_bitmap, &rl, 
+					nr_bm_clusters, nr_clusters);
 	else
-		enlarge_bitmap_data_attr(&rl, nr_bm_clusters, nr_clusters);
+		enlarge_bitmap_data_attr(vol, &resize->lcn_bitmap, &rl, 
+					 nr_bm_clusters, nr_clusters);
 
 	a->highest_vcn = cpu_to_le64(nr_bm_clusters - 1LL);
 	a->allocated_size = cpu_to_le64(nr_bm_clusters * vol->cluster_size);
 	a->data_size = cpu_to_le64(bm_bsize);
 	a->initialized_size = cpu_to_le64(bm_bsize);
 	
-	replace_attribute_runlist(resize->ctx, rl);
+	replace_attribute_runlist(vol, resize->ctx, rl);
 
 	/*
 	 * FIXME: update allocated/data sizes and timestamps in $FILE_NAME
 	 * attribute too, for now chkdsk will do this for us.
 	 */
 
-	size = ntfs_rl_pwrite(vol, rl, 0, bm_bsize, lcn_bitmap.bm);
+	size = ntfs_rl_pwrite(vol, rl, 0, bm_bsize, resize->lcn_bitmap.bm);
 	if (bm_bsize != size) {
 		if (size == -1)
 			perr_exit("Couldn't write $Bitmap");
@@ -1697,7 +1770,10 @@ static void truncate_bitmap_data_attr(ntfs_resize_t *resize)
  * Find the $DATA attribute (with or without a name) for the given MFT reference
  * (inode number).
  */
-static void lookup_data_attr(MFT_REF mref, const char *aname, ntfs_attr_search_ctx **ctx)
+static void lookup_data_attr(ntfs_volume *vol,
+			     MFT_REF mref, 
+			     const char *aname, 
+			     ntfs_attr_search_ctx **ctx)
 {
 	ntfs_inode *ni;
 	uchar_t *ustr = NULL;
@@ -1727,20 +1803,19 @@ static void lookup_data_attr(MFT_REF mref, const char *aname, ntfs_attr_search_c
  *
  * Shrink the $BadClus file to match the new volume size.
  */
-static void truncate_badclust_file(s64 nr_clusters)
+static void truncate_badclust_file(ntfs_resize_t *resize)
 {
-	ntfs_attr_search_ctx *ctx = NULL;
-
 	printf("Updating $BadClust file ...\n");
 
-	lookup_data_attr((MFT_REF)FILE_BadClus, "$Bad", &ctx);
+	lookup_data_attr(resize->vol, FILE_BadClus, "$Bad", &resize->ctx);
 	/* FIXME: sanity_check_attr(ctx->attr); */
-	truncate_badclust_bad_attr(ctx, nr_clusters);
+	truncate_badclust_bad_attr(resize);
 
-	if (write_mft_record(ctx->ntfs_ino->mft_no, ctx->mrec))
+	if (write_mft_record(resize->vol, resize->ctx->ntfs_ino->mft_no, 
+			     resize->ctx->mrec))
 		perr_exit("Couldn't update $BadClust");
 
-	ntfs_attr_put_search_ctx(ctx);
+	ntfs_attr_put_search_ctx(resize->ctx);
 }
 
 /**
@@ -1752,10 +1827,11 @@ static void truncate_bitmap_file(ntfs_resize_t *resize)
 {
 	printf("Updating $Bitmap file ...\n");
 
-	lookup_data_attr((MFT_REF)FILE_Bitmap, NULL, &resize->ctx);
+	lookup_data_attr(resize->vol, FILE_Bitmap, NULL, &resize->ctx);
 	truncate_bitmap_data_attr(resize);
 
-	if (write_mft_record(resize->ctx->ntfs_ino->mft_no, resize->ctx->mrec))
+	if (write_mft_record(resize->vol, resize->ctx->ntfs_ino->mft_no, 
+			     resize->ctx->mrec))
 		perr_exit("Couldn't update $Bitmap");
 
 	ntfs_attr_put_search_ctx(resize->ctx);
@@ -1768,15 +1844,15 @@ static void truncate_bitmap_file(ntfs_resize_t *resize)
  * All the bits are set to 0, except those representing the region beyond the
  * end of the disk.
  */
-static void setup_lcn_bitmap(void)
+static void setup_lcn_bitmap(struct bitmap *bm, s64 nr_clusters)
 {
 	/* Determine lcn bitmap byte size and allocate it. */
-	lcn_bitmap.size = nr_clusters_to_bitmap_byte_size(vol->nr_clusters);
+	bm->size = nr_clusters_to_bitmap_byte_size(nr_clusters);
 
-	if (!(lcn_bitmap.bm = (unsigned char *)calloc(1, lcn_bitmap.size)))
+	if (!(bm->bm = (unsigned char *)calloc(1, bm->size)))
 		perr_exit("Failed to allocate internal buffer");
 
-	bitmap_file_data_fixup(vol->nr_clusters, &lcn_bitmap);
+	bitmap_file_data_fixup(nr_clusters, bm);
 }
 
 /**
@@ -1788,6 +1864,7 @@ static void update_bootsector(ntfs_resize_t *r)
 {
 	NTFS_BOOT_SECTOR bs;
 	s64  bs_size = sizeof(NTFS_BOOT_SECTOR);
+	ntfs_volume *vol = r->vol;
 
 	printf("Updating Boot record ...\n");
 
@@ -1797,8 +1874,8 @@ static void update_bootsector(ntfs_resize_t *r)
 	if (vol->dev->d_ops->read(vol->dev, &bs, bs_size) == -1)
 		perr_exit("read() error");
 
-	bs.number_of_sectors = r->new_volume_size * bs.bpb.sectors_per_cluster;
-	bs.number_of_sectors = cpu_to_le64(bs.number_of_sectors);
+	bs.number_of_sectors = scpu_to_le64(r->new_volume_size *
+			bs.bpb.sectors_per_cluster);
 
 	if (r->mftmir_old) {
 		int save_progress = opt.show_progress;
@@ -1843,12 +1920,12 @@ static void print_vol_size(const char *str, s64 bytes)
  *
  * Display the amount of disk space in use.
  */
-static void print_disk_usage(ntfs_resize_t *resize)
+static void print_disk_usage(ntfs_volume *vol, s64 nr_used_clusters)
 {
 	s64 total, used;
 
 	total = vol->nr_clusters * vol->cluster_size;
-	used = resize->inuse * vol->cluster_size;
+	used = nr_used_clusters * vol->cluster_size;
 
 	/* WARNING: don't modify the text, external tools grep for it */
 	printf("Space in use       : %lld MB (%.1f%%)\n",
@@ -1858,7 +1935,7 @@ static void print_disk_usage(ntfs_resize_t *resize)
 
 static void print_num_of_relocations(ntfs_resize_t *resize)
 {
-	s64 relocations = resize->relocations * vol->cluster_size;
+	s64 relocations = resize->relocations * resize->vol->cluster_size;
 	
 	printf("Needed relocations : %lld (%lld MB)\n",
 			(long long)resize->relocations, (long long)
@@ -1872,9 +1949,10 @@ static void print_num_of_relocations(ntfs_resize_t *resize)
  * is dirty (Windows wasn't shutdown properly).  If everything is OK, then mount
  * the volume (load the metadata into memory).
  */
-static void mount_volume(void)
+static ntfs_volume *mount_volume(void)
 {
 	unsigned long mntflag;
+	ntfs_volume *vol = NULL;
 
 	if (ntfs_check_if_mounted(opt.volume, &mntflag)) {
 		perr_printf("Failed to check '%s' mount state", opt.volume);
@@ -1915,8 +1993,11 @@ static void mount_volume(void)
 	if (ntfs_version_is_supported(vol))
 		perr_exit("Unknown NTFS version");
 
-	printf("Cluster size       : %u bytes\n", vol->cluster_size);
+	printf("Cluster size       : %u bytes\n",
+			(unsigned int)vol->cluster_size);
 	print_vol_size("Current volume size", vol_size(vol, vol->nr_clusters));
+
+	return vol;
 }
 
 /**
@@ -1926,7 +2007,7 @@ static void mount_volume(void)
  * boots it will automatically run chkdsk to check for any problems.  If the
  * read-only command line option was given, this function will do nothing.
  */
-static void prepare_volume_fixup(void)
+static void prepare_volume_fixup(ntfs_volume *vol)
 {
 	u16 flags;
 
@@ -1970,7 +2051,7 @@ static void check_shrink_constraints(ntfs_resize_t *resize)
 	if (!resize->shrink)
 		return;
 
-	if (resize->inuse == vol->nr_clusters)
+	if (resize->inuse == resize->vol->nr_clusters)
 		err_exit("Volume is full. To shrink it, "
 			 "delete unused files.\n");
 	
@@ -1995,9 +2076,11 @@ static void check_shrink_constraints(ntfs_resize_t *resize)
 
 int main(int argc, char **argv)
 {
+	ntfs_fsck_t fsck;
 	ntfs_resize_t resize;
 	s64 new_size = 0;	/* in clusters; 0 = --info w/o --size */
 	s64 device_size;        /* in bytes */
+	ntfs_volume *vol;
 
 	printf("%s v%s\n", EXEC_NAME, VERSION);
 
@@ -2006,7 +2089,8 @@ int main(int argc, char **argv)
 
 	utils_set_locale();
 
-	mount_volume();
+	if ((vol = mount_volume()) == NULL)
+		err_exit("Couldn't open volume '%s'!\n", opt.volume);
 
 	device_size = ntfs_device_size_get(vol->dev, vol->sector_size);
 	device_size *= vol->sector_size;
@@ -2046,26 +2130,31 @@ int main(int argc, char **argv)
 		exit(0);
 	}
 
-	setup_lcn_bitmap();
-
-	memset(&resize, 0, sizeof(resize));
-	resize.new_volume_size = new_size;
-
-	/* This is also true if --info was used w/o --size (new_size = 0) */
-	if (new_size < vol->nr_clusters)
-		resize.shrink = 1;
+	memset(&fsck, 0, sizeof(fsck));
+	setup_lcn_bitmap(&fsck.lcn_bitmap, vol->nr_clusters);
 	
-	walk_inodes(&resize);
-	if (resize.multi_ref) {
+	build_allocation_bitmap(vol, &fsck);
+	if (fsck.multi_ref) {
 		err_printf("Filesystem check failed! Totally %d clusters "
-			   "referenced multiply times.\n", resize.multi_ref);
+			   "referenced multiply times.\n", fsck.multi_ref);
 		printf(corrupt_volume_msg);
 		exit(1);
 	}
-	compare_bitmaps(&lcn_bitmap);
+	compare_bitmaps(vol, &fsck.lcn_bitmap);
 
-	print_disk_usage(&resize);
+	print_disk_usage(vol, fsck.inuse);
 	
+	memset(&resize, 0, sizeof(resize));
+	resize.new_volume_size = new_size;
+	resize.inuse = fsck.inuse;
+	resize.lcn_bitmap = fsck.lcn_bitmap;
+	resize.vol = vol;
+	
+	/* This is also true if --info was used w/o --size (new_size = 0) */
+	if (new_size < vol->nr_clusters)
+		resize.shrink = 1;
+
+	set_resize_constrains(&resize);
 	set_disk_usage_constraint(&resize);
 	check_shrink_constraints(&resize);
 
@@ -2080,12 +2169,12 @@ int main(int argc, char **argv)
 	}
 
 	/* FIXME: performance - relocate logfile here if it's needed */
-	prepare_volume_fixup();
+	prepare_volume_fixup(vol);
 
 	if (resize.relocations)
 		relocate_inodes(&resize);
 
-	truncate_badclust_file(new_size);
+	truncate_badclust_file(&resize);
 	truncate_bitmap_file(&resize);
 	update_bootsector(&resize);
 
