@@ -72,7 +72,7 @@ GEN_PRINTF (Qprintf, stdout, &opts.quiet,   FALSE)
  *
  * Return:  none
  */
-void version (void)
+static void version (void)
 {
 	printf ("\n%s v%s - Recover deleted files from an NTFS Volume.\n\n",
 		EXEC_NAME, VERSION);
@@ -88,7 +88,7 @@ void version (void)
  *
  * Return:  none
  */
-void usage (void)
+static void usage (void)
 {
 	printf ("\nUsage: %s [options] device\n"
 		"    -s          --scan             Scan for files (default)\n"
@@ -135,7 +135,7 @@ void usage (void)
  * Return:  1, Success, the string was transformed
  *	    0, An error occurred
  */
-int transform (const char *pattern, char **regex)
+static int transform (const char *pattern, char **regex)
 {
 	char *result;
 	int length, i, j;
@@ -210,7 +210,7 @@ int transform (const char *pattern, char **regex)
  * Return:  1  Success
  *	    0  Error, the string was malformed
  */
-int parse_time (const char *value, time_t *since)
+static int parse_time (const char *value, time_t *since)
 {
 	time_t result, now;
 	char *suffix = NULL;
@@ -266,7 +266,7 @@ int parse_time (const char *value, time_t *since)
  * Return:  1 Success
  *	    0 Error, one or more problems
  */
-int parse_options (int argc, char *argv[])
+static int parse_options (int argc, char *argv[])
 {
 	static const char *sopt = "-b:Cc:d:fh?m:o:p:sS:t:u:qvV";
 	static const struct option lopt[] = {
@@ -327,7 +327,7 @@ int parse_options (int argc, char *argv[])
 			break;
 		case 'c':
 			if (opts.mode == MODE_NONE) {
-				if (!utils_parse_range (argv[optind-1],
+				if (!utils_parse_range (optarg,
 				    &opts.mft_begin, &opts.mft_end, TRUE))
 					err++;
 				opts.mode = MODE_COPY;
@@ -337,7 +337,7 @@ int parse_options (int argc, char *argv[])
 			break;
 		case 'd':
 			if (!opts.dest)
-				opts.dest = argv[optind-1];
+				opts.dest = optarg;
 			else
 				err++;
 			break;
@@ -350,7 +350,7 @@ int parse_options (int argc, char *argv[])
 			break;
 		case 'm':
 			if (!opts.match) {
-				if (!transform (argv[optind-1], &opts.match))
+				if (!transform (optarg, &opts.match))
 					err++;
 			} else {
 				err++;
@@ -358,7 +358,7 @@ int parse_options (int argc, char *argv[])
 			break;
 		case 'o':
 			if (!opts.output) {
-				opts.output = argv[optind-1];
+				opts.output = optarg;
 			} else {
 				err++;
 			}
@@ -384,14 +384,14 @@ int parse_options (int argc, char *argv[])
 			break;
 		case 'S':
 			if ((opts.size_begin > 0) || (opts.size_end > 0) ||
-			    !utils_parse_range (argv[optind-1], &opts.size_begin,
+			    !utils_parse_range (optarg, &opts.size_begin,
 			     &opts.size_end, TRUE)) {
 			    err++;
 			}
 			break;
 		case 't':
 			if (opts.since == 0) {
-				if (!parse_time (argv[optind-1], &opts.since))
+				if (!parse_time (optarg, &opts.since))
 					err++;
 			} else {
 			    err++;
@@ -515,7 +515,7 @@ int parse_options (int argc, char *argv[])
  *
  * Return:  none
  */
-void free_file (struct ufile *file)
+static void free_file (struct ufile *file)
 {
 	struct list_head *item, *tmp;
 
@@ -563,7 +563,7 @@ void free_file (struct ufile *file)
  * Return:  n  The number of $FILENAME attributes found
  *	   -1  Error
  */
-int get_filenames (struct ufile *file)
+static int get_filenames (struct ufile *file)
 {
 	ATTR_RECORD *rec;
 	FILE_NAME_ATTR *attr;
@@ -639,7 +639,7 @@ int get_filenames (struct ufile *file)
  * Return:  n  The number of $FILENAME attributes found
  *	   -1  Error
  */
-int get_data (struct ufile *file, ntfs_volume *vol)
+static int get_data (struct ufile *file, ntfs_volume *vol)
 {
 	ATTR_RECORD *rec;
 	ntfs_attr_search_ctx *ctx;
@@ -713,7 +713,7 @@ int get_data (struct ufile *file, ntfs_volume *vol)
  * Return:  Pointer  A ufile object containing the results
  *	    NULL     Error
  */
-struct ufile * read_record (ntfs_volume *vol, long long record)
+static struct ufile * read_record (ntfs_volume *vol, long long record)
 {
 	ATTR_RECORD *attr10, *attr20, *attr90;
 	struct ufile *file;
@@ -807,14 +807,14 @@ struct ufile * read_record (ntfs_volume *vol, long long record)
  * Return:  n  The percentage of the file that _could_ be recovered
  *	   -1  Error
  */
-int calc_percentage (struct ufile *file, ntfs_volume *vol)
+static int calc_percentage (struct ufile *file, ntfs_volume *vol)
 {
 	runlist_element *rl = NULL;
 	struct list_head *pos;
 	struct data *data;
 	long long i, j;
 	long long start, end;
-	int inuse, free;
+	int clusters_inuse, clusters_free;
 	int percent = 0;
 
 	if (!file || !vol)
@@ -832,8 +832,8 @@ int calc_percentage (struct ufile *file, ntfs_volume *vol)
 
 	list_for_each (pos, &file->data) {
 		data  = list_entry (pos, struct data, list);
-		inuse = 0;
-		free  = 0;
+		clusters_inuse = 0;
+		clusters_free  = 0;
 
 		if (data->encrypted) {
 			Vprintf ("File is encrypted, recovery is impossible.\n");
@@ -866,7 +866,7 @@ int calc_percentage (struct ufile *file, ntfs_volume *vol)
 		if (rl[0].lcn == LCN_RL_NOT_MAPPED) {	/* extended mft record */
 			Vprintf ("Missing segment at beginning, %lld "
 					"clusters\n", (long long)rl[0].length);
-			inuse += rl[0].length;
+			clusters_inuse += rl[0].length;
 			rl++;
 		}
 
@@ -875,12 +875,12 @@ int calc_percentage (struct ufile *file, ntfs_volume *vol)
 				Vprintf ("Missing segment at end, %lld "
 						"clusters\n",
 						(long long)rl[i].length);
-				inuse += rl[i].length;
+				clusters_inuse += rl[i].length;
 				continue;
 			}
 
 			if (rl[i].lcn == LCN_HOLE) {
-				free += rl[i].length;
+				clusters_free += rl[i].length;
 				continue;
 			}
 
@@ -889,18 +889,19 @@ int calc_percentage (struct ufile *file, ntfs_volume *vol)
 
 			for (j = start; j < end; j++) {
 				if (utils_cluster_in_use (vol, j))
-					inuse++;
+					clusters_inuse++;
 				else
-					free++;
+					clusters_free++;
 			}
 		}
-
-		if ((inuse + free) == 0) {
+		
+		if ((clusters_inuse + clusters_free) == 0) {
 			Eprintf ("Unexpected error whilst calculating percentage for inode %lld\n", file->inode);
 			continue;
 		}
 
-		data->percent = (free * 100) / (inuse + free);
+		data->percent = (clusters_free * 100) /
+				(clusters_inuse + clusters_free);
 
 		percent = max (percent, data->percent);
 	}
@@ -924,7 +925,7 @@ int calc_percentage (struct ufile *file, ntfs_volume *vol)
  *
  * Return:  none
  */
-void dump_record (struct ufile *file)
+static void dump_record (struct ufile *file)
 {
 	char buffer[20];
 	char *name;
@@ -1035,7 +1036,7 @@ void dump_record (struct ufile *file)
  *
  * Return:  none
  */
-void list_record (struct ufile *file)
+static void list_record (struct ufile *file)
 {
 	char buffer[20];
 	struct list_head *item;
@@ -1092,7 +1093,7 @@ void list_record (struct ufile *file)
  * Return:  1  There is a matching filename.
  *	    0  There is no match.
  */
-int name_match (regex_t *re, struct ufile *file)
+static int name_match (regex_t *re, struct ufile *file)
 {
 	struct list_head *item;
 	int result;
@@ -1130,7 +1131,8 @@ int name_match (regex_t *re, struct ufile *file)
  * Return:  -1  Error, something went wrong
  *	     0  Success, all the data was written
  */
-unsigned int write_data (int fd, const char *buffer, unsigned int bufsize)
+static unsigned int write_data (int fd, const char *buffer,
+	unsigned int bufsize)
 {
 	ssize_t result1, result2;
 
@@ -1177,8 +1179,8 @@ unsigned int write_data (int fd, const char *buffer, unsigned int bufsize)
  *
  * Return:  n  Length of the allocated name
  */
-int create_pathname (const char *dir, const char *name, const char *stream,
-		     char *buffer, int bufsize)
+static int create_pathname (const char *dir, const char *name, 
+	const char *stream, char *buffer, int bufsize)
 {
 	if (!name)
 		name = UNKNOWN;
@@ -1208,7 +1210,7 @@ int create_pathname (const char *dir, const char *name, const char *stream,
  * Return:  -1  Error, failed to create the file
  *	     n  Success, this is the file descriptor
  */
-int open_file (const char *pathname)
+static int open_file (const char *pathname)
 {
 	int flags;
 
@@ -1232,7 +1234,7 @@ int open_file (const char *pathname)
  * Return:  1  Success, set the file's date and time
  *	    0  Error, failed to change the file's date and time
  */
-int set_date (const char *pathname, time_t date)
+static int set_date (const char *pathname, time_t date)
 {
 	struct utimbuf ut;
 
@@ -1260,14 +1262,13 @@ int set_date (const char *pathname, time_t date)
  * Return:  -1  Error, something went wrong
  *	     n  Success, the number of recoverable files
  */
-int scan_disk (ntfs_volume *vol)
+static int scan_disk (ntfs_volume *vol)
 {
 	const int BUFSIZE = 8192;
 	char *buffer = NULL;
 	int results = 0;
 	ntfs_attr *attr;
 	long long size;
-	long long read;
 	long long bmpsize;
 	int i, j, k, b;
 	int percent;
@@ -1305,8 +1306,8 @@ int scan_disk (ntfs_volume *vol)
 	Qprintf ("Inode    Flags  %%age  Date            Size  Filename\n");
 	Qprintf ("---------------------------------------------------------------\n");
 	for (i = 0; i < bmpsize; i += BUFSIZE) {
-		read = min ((bmpsize - i), BUFSIZE);
-		size = ntfs_attr_pread (attr, i, read, buffer);
+		long long read_count = min ((bmpsize - i), BUFSIZE);
+		size = ntfs_attr_pread (attr, i, read_count, buffer);
 		if (size < 0)
 			break;
 
@@ -1383,7 +1384,7 @@ out:
  * Return:  0  Error, something went wrong
  *	    1  Success, the data was recovered
  */
-int undelete_file (ntfs_volume *vol, long long inode)
+static int undelete_file (ntfs_volume *vol, long long inode)
 {
 	char pathname[256];
 	char *buffer = NULL;
@@ -1586,7 +1587,7 @@ free:
  * Return:  0  Success, all the records were written
  *	    1  Error, something went wrong
  */
-int copy_mft (ntfs_volume *vol, long long mft_begin, long long mft_end)
+static int copy_mft (ntfs_volume *vol, long long mft_begin, long long mft_end)
 {
 	char pathname[256];
 	ntfs_attr *mft;
